unit ShellTree;

interface

uses
  Windows,  Messages,  SysUtils,  Classes,  Graphics,
  Controls,  Forms, StdCtrls,
  ExtCtrls,  ComCtrls,  ShellApi,  CommCtrl,
  ShlObj, ToolWin, ShellTreeTypes, ShellTreeProcs;

type
  TShellTree = class(TCustomTreeView)
  private
    FRootDir  : String;
    FRootPidl : TPidls;
    FOPtions  : TVisOptions;
    FRootUsePidl : Boolean;
    FDesignStart : Boolean;
    FInitialized : Boolean;
    procedure SetRootDir    (Value : String);
    procedure SetRootPidl   (Value : TPidls);
    procedure SetOptions    (Value : TVisOptions);
    procedure SetRootUsePidl(Value : Boolean);
    procedure SetDirectory  (Value : String);
    function  GetDirectory : String;
  protected
    function CreateNode : TTreeNode; override;
    procedure InitTree;
    procedure Loaded; override;
    function  CanExpand(Node: TTreeNode): Boolean; override;

    procedure CreateRoot;

    procedure EnumFolder(Node: TShellNode);
    procedure AddSubfolderToNode(Node: TShellNode; ItemIdList: PItemIdList);
    procedure MakeNodeInfo(Node : TShellNode; ParentNode: TShellNode; ShellFolder: IShellFolder;
                           ItemOfNode: PItemIdList);
  public
    Desktop, Root : IShellFolder;
    property Items stored false;
    procedure SelectPidl(ItemIdList : PItemIdList);
    function  GetPidlFromPath(Path : String; var Pidl : PItemIdList) : Boolean;
    constructor Create(AOwner : TComponent); override;
    destructor  Destroy; override;
    property Selected;
    procedure UpdateTree(LastAction : TLastAction);
  published
    property Align;
    property RootUsePIDL : Boolean read FRootUsePidl write SetRootUsePidl;
    property RootPIDL : TPidls read FRootPidl write SetRootPidl;
    property RootDir  : String read FRootDir  write SetRootDir;
    property Options  : TVisOptions read FOptions write SetOptions;
    property OnChange;
    property Directory : String read GetDirectory write SetDirectory;
  end;

procedure Register;

implementation

constructor TShelltree.Create(AOwner : TComponent);
begin
  inherited;
  FRootDir:=STD_DIR;
  FRootPidl:=PIDL_DESKTOP;
  FRootUsePidl:=True;
  FOptions:=[TF_FOLDERS];
  ReadOnly:=TRue;
  HideSelection:=False;
  FDesignStart:=False;
  FInitialized:=False;
//  IF NOT (csDesigning in ComponentState) Then begin
    SHGetMalloc(Allocator);
    SHGetDesktopFolder(Desktop)
//  end;
end;

procedure TShellTree.Loaded;
begin
  inherited;
//  IF NOT (csDesigning in ComponentState) Then begin
    UpdateTree(la_None);
//  end;
end;

destructor TShellTree.Destroy;
begin
//  IF Assigned(Desktop) Then Desktop._Release;
  inherited;
end;

procedure TShellTree.SelectPidl(ItemIdList : PItemIdList);
Var
  LongPidl,NodePidl : PItemIdList;
  LongLevel,NodeLevel : Integer;
  Node : TTreeNode;
  Found : Boolean;
begin
  LongPidl:=ItemIdList;
  Node:=Items.GetFirstNode;
  IF NOT Assigned(Node) Then begin
    Beep;
    exit;
  end;
  NodePidl:=(TShellNode(Node).NodeInfo.AbsoluteIDL);
  NodeLevel:=GetPIDLLevel(NodePidl);
  LongLevel :=GetPIDLLevel(ItemIDList);
  Found:=True;
  While (NodeLevel<LongLevel) AND Found DO begin
    Node.Expand(False);
    IF SearchForPidl(Node,LongPidl,NodeLevel) Then begin
      NodePidl:=(TShellNode(Node).NodeInfo.AbsoluteIDL);
      NodeLevel:=GetPIDLLevel(NodePidl);
    end else
      Found:=False;
  end;
  IF NodeLevel=LongLevel Then Selected:=Node;
end;

procedure TShellTree.UpdateTree(LastAction : TLastAction);
Var
  OldDir,OldRoot : String;
begin
  IF NOT FInitialized Then InitTree;
  OldDir:=Directory;
  OldRoot:=RootDir;
  IF Not (csloading in ComponentState) Then
    CreateRoot;
  IF (LastAction=la_RootDir) Then Directory:=OldDir;
  IF (Length(OldRoot)>1) AND (LastAction<>la_RootDir) Then RootDir:=OldRoot;
end;

function TShellTree.GetPidlFromPath(Path : String; var Pidl : PItemIdList) : Boolean;
Var
  WCA : ARRAY[0..300] OF WideChar;
  S : Integer;
  UL : ULong;
begin
  Path:=TestDirectory(Path);
  S:=Length(Path);
  StringToWideChar(Path,WCA,SizeOf(WCA));
  RESULT:=(Desktop.ParseDisplayName(handle,nil,@WCA,S,Pidl,UL)=NOERROR);
end;

function TShellTree.CanExpand(Node: TTreeNode): Boolean;
Var
  SNode : TShellNode;
begin
  SNode:=TShellNode(Node);
  if not SNode.NodeInfo.Expanded then begin
    EnumFolder(SNode);
    SNode.NodeInfo.Expanded:=True;
  end;
  Result:=((NOT (csDesigning in ComponentState) OR
    ((csDesigning in ComponentState) AND FDesignStart)));
end;

procedure TShellTree.EnumFolder(Node: TShellNode);
var
  ShellFolder: IShellFolder;
  Objects: IEnumIdList;
  ItemIdList: PItemIdList;
  DummyResult: ULONG;
  Count: integer;
  Flags : DWord;
begin
  Count:=0;
  ShellFolder:=Node.NodeInfo.ShellFolder;
  Flags:=0;
  IF Node.NodeInfo.RelativeIDL<>nil Then
    IF TF_Folders    in Options Then Flags:=Flags + SHCONTF_FOLDERS;
  IF TF_NonFolders in Options Then Flags:=Flags + SHCONTF_NONFOLDERS;
  IF TF_Hidden     in Options Then Flags:=Flags + SHCONTF_INCLUDEHIDDEN;
  IF Succeeded(ShellFolder.EnumObjects(Handle, Flags, Objects)) then begin
    while Objects.Next(1, ItemIdList, DummyResult)=NOERROR do begin
      AddSubfolderToNode(Node, ItemIdList);
      inc(Count);
    end;
    if Count=0 then Node.HasChildren:=False;
  end;
  Node.CustomSort(@ShelltreeSortProc, 0);
end;

procedure TShellTree.AddSubfolderToNode(Node: TShellNode; ItemIdList: PItemIdList);
var
  ShellFolder: IShellFolder;
  DisplayName: string;
  TreeNode: TShellNode;
  FileInfo: TSHFileInfo;
begin
  ShellFolder:=Node.NodeInfo.ShellFolder;
  DisplayName:=GetShellItemName(ShellFolder, ItemIdList);
  TreeNode:=TShellNode(Node.Owner.AddChild(Node, DisplayName));
  MakeNodeInfo(TreeNode,Node, nil, ItemIdList);
  TreeNode.HasChildren:=HasSubfolders(ShellFolder, ItemIdList);
  ShellFolder.BindToObject(ItemIdList, nil, IID_ISHELLFOLDER,pointer(TreeNode.NodeInfo.ShellFolder));
  SHGetFileInfo(PChar(TreeNode.NodeInfo.AbsoluteIDL), 0, FileInfo, sizeof(FileInfo),SHGFI_PIDL or SHGFI_SYSICONINDEX);
  TreeNode.ImageIndex:=FileInfo.iIcon;
  SHGetFileInfo(PChar(TreeNode.NodeInfo.AbsoluteIDL), 0, FileInfo, sizeof(FileInfo),SHGFI_PIDL or SHGFI_SYSICONINDEX or SHGFI_OPENICON);
  TreeNode.SelectedIndex:=FileInfo.iIcon;
end;

function TSHellTree.CreateNode : TTreeNode;
begin
  Result:=TShellNode.Create(Items);
end;

procedure TShellTree.InitTree;
var
  FileInfo : TSHFileInfo;
  DesktopItemIdList : PItemIdList;
  SysImageListSmall : HImageList;
begin
  SHGetSpecialFolderLocation(Handle, CSIDL_DESKTOP, DesktopItemIdList);
  SysImageListSmall:=SHGetFileInfo(PChar(DesktopItemIdList), 0, FileInfo, sizeof(FileInfo),
                                   SHGFI_PIDL or SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
  FreePidl(DesktopItemIdList);
  TreeView_SetImageList(Handle, SysImageListSmall, TVSIL_NORMAL);
  FInitialized:=True;
end;

procedure TShellTree.CreateRoot;
Var
  FileInfo : TSHFileInfo;
  IdList : PItemIdList;
  Node : TShellNode;
begin
  //IF (csDesigning in ComponentState) Then exit;
  IF RootUsePidl Then begin
    IF NOT Succeeded(SHGetSpecialFolderLocation(Handle,ConvertPidl(RootPidl), IdList)) then begin
      SHGetSpecialFolderLocation(Handle,CSIDL_DESKTOP, IdList);
    end;
  end else begin
    IF NOT GetPidlFromPath(RootDir,IdList) Then
      GetPidlFromPath(STD_DIR,IdList);
  end;
  Items.Clear;
  SHGetFileInfo(PChar(IdList), 0, FileInfo, sizeof(FileInfo),SHGFI_PIDL or SHGFI_SYSICONINDEX or SHGFI_SMALLICON);
  Node:=TShellNode(Items.Add(Nil,''));
  Node.ImageIndex:=FileInfo.iIcon;
  Node.SelectedIndex:=FileInfo.iIcon;
  Root:=GetIShellFolder(Desktop,IdList);
  MakeNodeInfo(Node,nil, Root, IDList);
  Node.Text:=GetShellItemName(Desktop,IDList);
  Node.HasChildren:=True;
  FDesignStart:=True;
  Node.Expand(False);
  Selected:=Node;
  IF NOT RootUsePidl Then FRootDir:=GetPathFromPidl(IDList);
end;

procedure TShellTree.SetRootDir(Value : String);
begin
  Value:=TestDirectory(Value);
  IF Value<>FRootDir Then begin
    FRootDir:=Value;
    IF NOT RootUsePidl Then UpdateTree(la_RootDir);
  end;
end;

procedure TShellTree.SetRootPidl(Value : TPidls);
begin
  IF Value<>FRootPidl Then begin
    FRootPidl:=Value;
    IF RootUsePidl Then UpdateTree(la_RootPidl);
  end;
end;

procedure TShellTree.SetOptions (Value : TVisOptions);
begin
  IF Value <> FOPtions Then begin
    FOptions :=Value;
    UpdateTree(la_None);
  end;
end;

procedure TShellTree.SetDirectory  (Value : String);
Var
  Pidl : PItemIdList;
begin
  Value:=TestDirectory(Value);
  IF Value<>GetDirectory Then begin
    IF not (csreading in ComponentState) Then begin
      GetPidlFromPath(Value,Pidl);
      SelectPidl(Pidl);
      FreePidl(Pidl);
    end;
  end;
end;

function TShelltree.GetDirectory : String;
Var
  SN : TSHellNode;
begin
  IF Assigned(Selected) Then begin
    SN:=TShellNode(Selected);
    IF GetPidlLevel(SN.NodeInfo.AbsoluteIDL)>0 Then
      Result:=GetPathFromPIDL(SN.NodeInfo.AbsoluteIDL)
    else Result:='';
  end else
    Result:='';
end;

procedure TShellTree.SetRootUsePidl(Value : Boolean);
begin
  IF Value<>FRootUsePidl Then begin
    FRootUsePidl:=Value;
    UpdateTree(la_None);
  end;
end;

procedure TShellTree.MakeNodeInfo(Node : TShellNode; ParentNode: TShellNode;
                      ShellFolder: IShellFolder;
                      ItemOfNode: PItemIdList);
begin
  FillChar(Node.NodeInfo, sizeof(Node.Nodeinfo), 0);
  With Node DO begin
    Nodeinfo.ShellFolder:=ShellFolder;
    Nodeinfo.RelativeIDL:=ItemOfNode;
    IF Assigned(Node.Parent) Then
      Nodeinfo.AbsoluteIDL:=MakeAbsoluteIDL(ParentNode, ItemOfNode)
    else
      Nodeinfo.AbsoluteIDL:=ItemOfNode;
    Nodeinfo.Expanded:=False;
  end;
end;

procedure Register;
begin
  RegisterComponents('Samples', [TShellTree]);
end;

end.
