unit ShellTreeProcs;

interface

Uses
  Windows, Comctrls, ShlObj, ShellTreetypes, SysUtils, FileCtrl;

function  GetPathLen(Path: PItemIdList): Integer;
function  GetShellItemName(Folder: IShellFolder; ItemIdList: PItemIdList): string;
function  GetPIDLLevel(Path : PItemIdList) : Integer;
function  GetPathFromPidl(IdList : PItemIDList) : String;
function  ConvertPidl(P : TPidls) : Integer;
function  GetIShellFolder(Desktop : IShellFolder;IdList : PItemIDList) : ISHellFolder;
function  IsShellItemFromFileSystem(Folder: IShellFolder; ItemIdList: PItemIdList): Boolean;
function  ShellTreeSortProc(Node1, Node2: TTreeNode; Data: integer): integer; stdcall;
function  TestDirectory(Dir : String) : String;
function  MakeAbsoluteIDL(ParentNode: TShellNode; Child: PItemIdList): PItemIdList;
function  HasSubfolders(Folder: IShellFolder; ItemIdList: PItemIdList): Boolean;
function  ComparePidl(Level : Integer;LongPidl,Pidl : PItemIdList) : Boolean;
function  SearchForPidl(Var Node : TTreeNode;Pidl : PItemIdList;Level : Integer): Boolean;
procedure FreePidl(Pidl : PItemIdList);

implementation

procedure FreePidl(Pidl : PItemIdList);
begin
  if (Assigned(Pidl)) then Allocator.Free(Pidl)
end;

function SearchForPidl(Var Node : TTreeNode;Pidl : PItemIdList;Level : Integer): Boolean;
Var
  CNode : TShellNode;
  NodePidl : PItemIdList;
begin
  CNode:=TShellNode(Node.GetFirstChild);
  Result:=False;
  While (CNode<>nil) AND NOt Result DO begin
    NodePidl:=CNode.NodeInfo.RelativeIDL;
    Result:=ComparePidl(Level,Pidl,NodePidl);
    IF NOT Result Then CNode:=TShellNode(Node.GetNextChild(TTreeNode(CNode)));
  end;
  IF Result Then Node:=CNode;
end;

function ComparePidl(Level : Integer;LongPidl,Pidl : PItemIdList) : Boolean;
Var
  X : Integer;
begin
  Result:=False;
  IF Level>GetPidlLevel(LongPidl) Then exit;
  FOR X:=1 TO Level DO LongPidl:=PItemIdList(@LongPidl^.mkId.abID[LongPidl^.mkId.cb-2]);
  IF LongPidl^.mkid.cb=Pidl^.mkid.cb Then begin
    Result:=True;
    FOR X:=0 TO (LongPidl^.mkid.cb-3) DO begin
      IF LongPidl^.mkid.abid[X]<>Pidl^.mkid.abid[X] Then Result:=False;
    end;
  end;
end;

function TestDirectory(Dir : String) : String;
begin
  Result:=Dir;
  While (Length(Result)>3) AND NOT DirectoryExists(Result) DO
    Result:=ExtractFilePath(Result);
  IF NOT DirectoryExists(Result) Then Result:=STD_DIR;
  IF (Length(Result)>3) AND (Result[Length(Result)]='\') Then
    Result:=Copy(Result,1,Length(Result)-1);
end;

function GetIShellFolder(Desktop : IShellFolder;IdList : PItemIDList) : ISHellFolder;
Var
  X : Integer;
begin
  Result:=Desktop;
  IF GetPIDLLevel(IDList)>0 Then begin
    Result._Release;
    Result.BindToObject(IdList, nil, IID_ISHELLFOLDER, pointer(Result));
  end;
end;

function ConvertPidl(P : TPidls) : Integer;
begin
  Case P OF
     PIDL_BITBUCKET                : Result := CSIDL_BITBUCKET;
     PIDL_CONTROLS                 : Result := CSIDL_CONTROLS;
     PIDL_DESKTOP                  : Result := CSIDL_DESKTOP;
     PIDL_DESKTOPDIRECTORY         : Result := CSIDL_DESKTOPDIRECTORY;
     PIDL_DRIVES                   : Result := CSIDL_DRIVES;
     PIDL_FONTS                    : Result := CSIDL_FONTS;
     PIDL_NETHOOD                  : Result := CSIDL_NETHOOD;
     PIDL_NETWORK                  : Result := CSIDL_NETWORK;
     PIDL_PERSONAL                 : Result := CSIDL_PERSONAL;
     PIDL_PRINTERS                 : Result := CSIDL_PRINTERS;
     PIDL_PROGRAMS                 : Result := CSIDL_PROGRAMS;
     PIDL_RECENT                   : Result := CSIDL_RECENT;
     PIDL_SENDTO                   : Result := CSIDL_SENDTO;
     PIDL_STARTMENU                : Result := CSIDL_STARTMENU;
     PIDL_STARTUP                  : Result := CSIDL_STARTUP;
     PIDL_TEMPLATES                : Result := CSIDL_TEMPLATES;
     PIDL_FAVORITES                : Result := CSIDL_FAVORITES;
  end;
end;

function  ShellTreeSortProc(Node1, Node2: TTreeNode; Data: integer): integer; stdcall;
Var
  ISF : IShellFolder;
  N1,N2 : TShellNode;
  Res : HResult;
begin
  IF Assigned(Node1.Parent) Then begin
    N1:=TShellNode(Node1);
    N2:=TShellNode(Node2);
    ISF:=TShellNode(Node1.parent).NodeInfo.ShellFolder;
    Res:=ISF.CompareIDs(0,N1.NodeInfo.RelativeIDL,N2.NodeInfo.RelativeIDL);
    IF (Res<>0) AND (Res<>1) Then Res:=-1;
    Result:=Res;
  end;
end;

function GetPathFromPidl(IdList : PItemIDList) : String;
Var
  S : String;
  I : Integer;
begin
  SetString(S,PChar(nil),MAX_PATH+1);
  SHGetPathFromIDList(IdList,Pointer(S));
  I:=Pos(#0,S);
  IF I>0 Then Result:=Copy(S,1,I-1)
    else Result:=S;
end;

function GetPIDLLevel(Path : PItemIdList) : Integer;
begin
  Result:=0;
  if not Assigned(Path) then exit;
  while (Path^.mkId.cb<>0) AND (Result<200) do begin
    inc(Result);
    Path:=PItemIdList(@Path^.mkId.abID[Path^.mkId.cb-2]);
  end;
end;

function IsShellItemFromFileSystem(Folder: IShellFolder; ItemIdList: PItemIdList): Boolean;
var
  Attrib: UINT;
begin
  Attrib:=SFGAO_FileSystem;
  Folder.GetAttributesOf(1, ItemIdList, Attrib);
  Result:=Attrib=(Attrib or SFGAO_FileSystem);
end;

function HasSubfolders(Folder: IShellFolder; ItemIdList: PItemIdList): Boolean;
var
  Attrib: UINT;
begin
  Attrib:=SFGAO_HasSubFolder;
  Folder.GetAttributesOf(1, ItemIdList, Attrib);
  Result:=Attrib=(Attrib or SFGAO_HasSubFolder);
end;

function GetShellItemName(Folder: IShellFolder; ItemIdList: PItemIdList): string;
var
  StrResult: TStrRet;
  DisplayName: PChar;
begin
  Folder.GetDisplayNameOf(ItemIdList, 0, StrResult);
  case StrResult.uType of
    0: DisplayName:='OLE-String! Kann hier nicht gelesen werden.';
    1: DisplayName:=PChar(ItemIdList)+StrResult.uOffset;
    2: DisplayName:=StrResult.cStr;
  end;
  Result:=DisplayName;
end;

function GetPathLen(Path: PItemIdList): Integer;
begin
  Result:=0;
  if not Assigned(Path) then exit;
  while Path^.mkId.cb<>0 do begin
    inc(Result, Path^.mkId.cb);
    Path:=PItemIdList(@Path^.mkId.abID[Path^.mkId.cb-2]);
  end;
end;

function MakeAbsoluteIDL(ParentNode: TShellNode; Child: PItemIdList): PItemIdList;
var
  PathLen: integer;
  ParentFolder: IShellFolder;
  ParentPath: PItemIdList;
begin
  ParentFolder:=nil;
  if Assigned(ParentNode) then begin
    ParentFolder:=ParentNode.NodeInfo.ShellFolder;
    ParentPath  :=ParentNode.NodeInfo.AbsoluteIDL;
  end;
  if Assigned(ParentFolder) then
    PathLen:=GetPathLen(ParentPath)
  else PathLen:=0;
  Result:=Allocator.Alloc(PathLen+Child^.mkId.cb+2); // +2 fr die Endemarkierung
  if PathLen<>0 then begin
    system.move(ParentPath^, Result^, PathLen);
    system.move(Child^, Result^.mkId.abID[PathLen-2], Child^.mkId.cb);
  end else
    system.move(Child^, Result^, Child^.mkId.cb);
  Result^.mkId.abID[PathLen+Child^.mkId.cb-2]:=0;
  Result^.mkId.abID[PathLen+Child^.mkId.cb-2+1]:=0;
end;

end.
